# Copyright 2021-2024 Nitrokey Developers
#
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.

from typing import List, Optional

from ._base import Model as Model  # noqa: F401
from ._base import TrussedBase as TrussedBase  # noqa: F401
from ._bootloader import FirmwareContainer as FirmwareContainer  # noqa: F401
from ._bootloader import FirmwareMetadata as FirmwareMetadata  # noqa: F401
from ._bootloader import TrussedBootloader as TrussedBootloader  # noqa: F401
from ._bootloader import Variant as Variant  # noqa: F401
from ._bootloader import parse_firmware_image as parse_firmware_image  # noqa: F401
from ._device import App as App  # noqa: F401
from ._device import TrussedDevice as TrussedDevice  # noqa: F401
from ._exceptions import TimeoutException as TimeoutException  # noqa: F401
from ._exceptions import TrussedException as TrussedException  # noqa: F401
from ._utils import Fido2Certs as Fido2Certs  # noqa: F401
from ._utils import Uuid as Uuid  # noqa: F401
from ._utils import Version as Version  # noqa: F401


def list(*, model: Optional[Model] = None) -> List[TrussedBase]:
    devices: List[TrussedBase] = []

    if model is None or model == Model.NK3:
        from nitrokey import nk3

        devices.extend(nk3.list())

    if model is None or model == Model.NKPK:
        from nitrokey import nkpk

        devices.extend(nkpk.list())

    return devices


def open(path: str, *, model: Optional[Model] = None) -> Optional[TrussedBase]:
    devices: List[TrussedBase] = []

    if model is None or model == Model.NK3:
        from nitrokey import nk3

        nk3_device = nk3.open(path)
        if nk3_device is not None:
            devices.append(nk3_device)

    if model is None or model == Model.NKPK:
        from nitrokey import nkpk

        nkpk_device = nkpk.open(path)
        if nkpk_device is not None:
            devices.append(nkpk_device)

    if len(devices) > 1:
        raise Exception(f"Found multiple devices at path {path}")
    if len(devices) == 1:
        return devices[0]
    return None



# Copyright 2021-2022 Nitrokey Developers
#
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.

from typing import List, Optional, Union

from nitrokey.trussed._bootloader import ModelData
from nitrokey.trussed._bootloader.nrf52 import SignatureKey

from ._bootloader import NK3Bootloader as NK3Bootloader  # noqa: F401
from ._device import NK3 as NK3  # noqa: F401

_PID_NK3_DEVICE = 0x42B2
_PID_NK3_LPC55_BOOTLOADER = 0x42DD
_PID_NK3_NRF52_BOOTLOADER = 0x42E8

_NK3_DATA = ModelData(
    firmware_repository_name="nitrokey-3-firmware",
    firmware_pattern_string="firmware-nk3-v.*\\.zip$",
    nrf52_signature_keys=[
        SignatureKey(
            name="Nitrokey",
            is_official=True,
            der="3059301306072a8648ce3d020106082a8648ce3d03010703420004a0849b19007ccd4661c01c533804b7fd0c4d8c0e7583653f1f36a8331afff298b542bd00a3dc47c16bf428ac4d2864137d63f702d89e5b42674e0549b4232618",
        ),
        SignatureKey(
            name="Nitrokey Test",
            is_official=False,
            der="3059301306072a8648ce3d020106082a8648ce3d0301070342000493e461ab0582bda1f45b0ce47d66bc4e8623e289c31af2098cde6ebd8631da85acf17e412d406c1e38c2de654a8fd0196506a85b169a756aeac2505a541cdd5d",
        ),
    ],
)


def list() -> List[Union[NK3, NK3Bootloader]]:
    devices: List[Union[NK3, NK3Bootloader]] = []
    devices.extend(NK3Bootloader.list())
    devices.extend(NK3.list())
    return devices


def open(path: str) -> Optional[Union[NK3, NK3Bootloader]]:
    device = NK3.open(path)
    bootloader_device = NK3Bootloader.open(path)
    if device and bootloader_device:
        raise Exception(f"Found multiple devices at path {path}")
    if device:
        return device
    if bootloader_device:
        return bootloader_device
    return None



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Based crypto classes."""
from typing import Dict

from cryptography import utils
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.x509.base import Version
from cryptography.x509.extensions import Extensions
from cryptography.x509.name import Name
from cryptography.x509.oid import ExtensionOID

from ..exceptions import SPSDKError


class SPSDKEncoding(utils.Enum):
    """Extension of cryptography Encoders class."""

    NXP = "NXP"
    PEM = "PEM"
    DER = "DER"

    @staticmethod
    def get_cryptography_encodings(encoding: "SPSDKEncoding") -> Encoding:
        """Get Encoding in cryptography class."""
        cryptography_encoding = {
            SPSDKEncoding.PEM: Encoding.PEM,
            SPSDKEncoding.DER: Encoding.DER,
        }.get(encoding)
        if cryptography_encoding is None:
            raise SPSDKError(f"{encoding} format is not supported by cryptography.")
        return cryptography_encoding

    @staticmethod
    def get_file_encodings(data: bytes) -> "SPSDKEncoding":
        """Get the encoding type out of given item from the data.

        :param data: Already loaded data file to determine the encoding style
        :return: encoding type (Encoding.PEM, Encoding.DER)
        """
        encoding = SPSDKEncoding.PEM
        try:
            decoded = data.decode("utf-8")
        except UnicodeDecodeError:
            encoding = SPSDKEncoding.DER
        else:
            if decoded.find("----") == -1:
                encoding = SPSDKEncoding.DER
        return encoding

    @staticmethod
    def all() -> Dict[str, "SPSDKEncoding"]:
        """Get all supported encodings."""
        return {
            "NXP": SPSDKEncoding.NXP,
            "PEM": SPSDKEncoding.PEM,
            "DER": SPSDKEncoding.DER,
        }


SPSDKExtensions = Extensions
SPSDKExtensionOID = ExtensionOID
SPSDKName = Name
SPSDKVersion = Version



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Low level device base class."""
from abc import ABC, abstractmethod
from types import TracebackType
from typing import TYPE_CHECKING, Optional, Type

if TYPE_CHECKING:
    from typing_extensions import Self


class DeviceBase(ABC):
    """Device base class."""

    def __enter__(self) -> "Self":
        self.open()
        return self

    def __exit__(
        self,
        exception_type: Optional[Type[Exception]] = None,
        exception_value: Optional[Exception] = None,
        traceback: Optional[TracebackType] = None,
    ) -> None:
        self.close()

    @property
    @abstractmethod
    def is_opened(self) -> bool:
        """Indicates whether interface is open."""

    @abstractmethod
    def open(self) -> None:
        """Open the interface."""

    @abstractmethod
    def close(self) -> None:
        """Close the interface."""

    @abstractmethod
    def read(self, length: int, timeout: Optional[int] = None) -> bytes:
        """Read data from the device.

        :param length: Length of data to be read
        :param timeout: Read timeout to be applied
        """

    @abstractmethod
    def write(self, data: bytes, timeout: Optional[int] = None) -> None:
        """Write data to the device.

        :param data: Data to be written
        :param timeout: Read timeout to be applied
        """

    @property
    @abstractmethod
    def timeout(self) -> int:
        """Timeout property."""

    @timeout.setter
    @abstractmethod
    def timeout(self, value: int) -> None:
        """Timeout property setter."""

    @abstractmethod
    def __str__(self) -> str:
        """Return string containing information about the interface."""



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2024 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""OpenSSL implementation for symmetric key encryption."""


# Used security modules
from typing import Optional

from cryptography.hazmat.primitives import keywrap
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from ..utils.misc import Endianness


class Counter:
    """AES counter with specified counter byte ordering and customizable increment."""

    @property
    def value(self) -> bytes:
        """Initial vector for AES encryption."""
        return self._nonce + self._ctr.to_bytes(4, self._ctr_byteorder_encoding.value)

    def __init__(
        self,
        nonce: bytes,
        ctr_value: Optional[int] = None,
        ctr_byteorder_encoding: Endianness = Endianness.LITTLE,
    ):
        """Constructor.

        :param nonce: last four bytes are used as initial value for counter
        :param ctr_value: counter initial value; it is added to counter value retrieved from nonce
        :param ctr_byteorder_encoding: way how the counter is encoded into output value
        :raises SPSDKError: When invalid byteorder is provided
        """
        assert isinstance(nonce, bytes) and len(nonce) == 16
        self._nonce = nonce[:-4]
        self._ctr_byteorder_encoding = ctr_byteorder_encoding
        self._ctr = int.from_bytes(nonce[-4:], ctr_byteorder_encoding.value)
        if ctr_value is not None:
            self._ctr += ctr_value

    def increment(self, value: int = 1) -> None:
        """Increment counter by specified value.

        :param value: to add to counter
        """
        self._ctr += value


def aes_key_unwrap(kek: bytes, wrapped_key: bytes) -> bytes:
    """Unwraps a key using a key-encrypting key (KEK).

    :param kek: The key-encrypting key
    :param wrapped_key: Encrypted data
    :return: Un-wrapped key
    """
    return keywrap.aes_key_unwrap(kek, wrapped_key)


def aes_ctr_decrypt(key: bytes, encrypted_data: bytes, nonce: bytes) -> bytes:
    """Decrypt encrypted data with AES in CTR mode.

    :param key: The key for data decryption
    :param encrypted_data: Input data
    :param nonce: Nonce data with counter value
    :return: Decrypted data
    """
    cipher = Cipher(algorithms.AES(key), modes.CTR(nonce))
    enc = cipher.decryptor()
    return enc.update(encrypted_data) + enc.finalize()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module provides exceptions for SPSDK utilities."""
from ..exceptions import SPSDKError


class SPSDKTimeoutError(TimeoutError, SPSDKError):
    """SPSDK Timeout."""



# Copyright 2021 Nitrokey Developers
#
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.

from typing import List

from fido2.hid import CtapHidDevice

from nitrokey import _VID_NITROKEY
from nitrokey.trussed import Fido2Certs, Model, TrussedDevice, Version

FIDO2_CERTS = [
    Fido2Certs(
        start=Version(0, 1, 0),
        hashes=[
            "ad8fd1d16f59104b9e06ef323cc03f777ed5303cd421a101c9cb00bb3fdf722d",
        ],
    ),
    Fido2Certs(
        start=Version(1, 0, 3),
        hashes=[
            "aa1cb760c2879530e7d7fed3da75345d25774be9cfdbbcbd36fdee767025f34b",  # NK3xN/lpc55
            "4c331d7af869fd1d8217198b917a33d1fa503e9778da7638504a64a438661ae0",  # NK3AM/nrf52
            "f1ed1aba24b16e8e3fabcda72b10cbfa54488d3b778bda552162d60c6dd7b4fa",  # NK3AM/nrf52 test
        ],
    ),
]


class NK3(TrussedDevice):
    """A Nitrokey 3 device running the firmware."""

    def __init__(self, device: CtapHidDevice) -> None:
        super().__init__(device, FIDO2_CERTS)

    @property
    def model(self) -> Model:
        return Model.NK3

    @property
    def pid(self) -> int:
        from . import _PID_NK3_DEVICE

        return _PID_NK3_DEVICE

    @property
    def name(self) -> str:
        return "Nitrokey 3"

    @classmethod
    def from_device(cls, device: CtapHidDevice) -> "NK3":
        return cls(device)

    @classmethod
    def list(cls) -> List["NK3"]:
        from . import _PID_NK3_DEVICE

        return cls._list_vid_pid(_VID_NITROKEY, _PID_NK3_DEVICE)



#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from . import dfu_cc_pb2 as pb


class InitPacketPB:
    def __init__(self, from_bytes: bytes) -> None:
        # construct from a protobuf string/buffer
        self.packet = pb.Packet()
        self.packet.ParseFromString(from_bytes)

        if self.packet.HasField("signed_command"):
            self.init_command = self.packet.signed_command.command.init
        else:
            self.init_command = self.packet.command.init

        self._validate()

    def _validate(self) -> None:
        if (
            self.init_command.type == pb.APPLICATION
            or self.init_command.type == pb.EXTERNAL_APPLICATION
        ) and self.init_command.app_size == 0:
            raise RuntimeError(
                "app_size is not set. It must be set when type is APPLICATION/EXTERNAL_APPLICATION"
            )
        elif self.init_command.type == pb.SOFTDEVICE and self.init_command.sd_size == 0:
            raise RuntimeError(
                "sd_size is not set. It must be set when type is SOFTDEVICE"
            )
        elif self.init_command.type == pb.BOOTLOADER and self.init_command.bl_size == 0:
            raise RuntimeError(
                "bl_size is not set. It must be set when type is BOOTLOADER"
            )
        elif self.init_command.type == pb.SOFTDEVICE_BOOTLOADER and (
            self.init_command.sd_size == 0 or self.init_command.bl_size == 0
        ):
            raise RuntimeError(
                "Either sd_size or bl_size is not set. Both must be set when type "
                "is SOFTDEVICE_BOOTLOADER"
            )

        if (
            self.init_command.fw_version < 0
            or self.init_command.fw_version > 0xFFFFFFFF
            or self.init_command.hw_version < 0
            or self.init_command.hw_version > 0xFFFFFFFF
        ):
            raise RuntimeError(
                "Invalid range of firmware argument. [0 - 0xffffffff] is valid range"
            )

    def get_init_command_bytes(self) -> bytes:
        return self.init_command.SerializeToString()

    def __str__(self) -> str:
        return str(self.init_command)



import time


class Try:
    """Utility class for an execution of a repeated action with Retries."""

    def __init__(self, i: int, retries: int) -> None:
        self.i = i
        self.retries = retries

    def __str__(self) -> str:
        return f"try {self.i + 1} of {self.retries}"

    def __repr__(self) -> str:
        return f"Try(i={self.i}, retries={self.retries})"


class Retries:
    """Utility class for repeating an action multiple times until it succeeds."""

    def __init__(self, retries: int, timeout: float = 0.5) -> None:
        self.retries = retries
        self.i = 0
        self.timeout = timeout

    def __iter__(self) -> "Retries":
        return self

    def __next__(self) -> Try:
        if self.i >= self.retries:
            raise StopIteration
        if self.i > 0:
            time.sleep(self.timeout)
        t = Try(self.i, self.retries)
        self.i += 1
        return t



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/nitrokey/trussed/_bootloader/nrf52_upload/dfu/dfu-cc.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n>src/nitrokey/trussed/_bootloader/nrf52_upload/dfu/dfu-cc.proto\x12\x03\x64\x66u"6\n\x04Hash\x12 \n\thash_type\x18\x01 \x02(\x0e\x32\r.dfu.HashType\x12\x0c\n\x04hash\x18\x02 \x02(\x0c"C\n\x0e\x42ootValidation\x12!\n\x04type\x18\x01 \x02(\x0e\x32\x13.dfu.ValidationType\x12\x0e\n\x06\x62ytes_\x18\x02 \x02(\x0c"\xf8\x01\n\x0bInitCommand\x12\x12\n\nfw_version\x18\x01 \x01(\r\x12\x12\n\nhw_version\x18\x02 \x01(\r\x12\x12\n\x06sd_req\x18\x03 \x03(\rB\x02\x10\x01\x12\x19\n\x04type\x18\x04 \x01(\x0e\x32\x0b.dfu.FwType\x12\x0f\n\x07sd_size\x18\x05 \x01(\r\x12\x0f\n\x07\x62l_size\x18\x06 \x01(\r\x12\x10\n\x08\x61pp_size\x18\x07 \x01(\r\x12\x17\n\x04hash\x18\x08 \x01(\x0b\x32\t.dfu.Hash\x12\x17\n\x08is_debug\x18\t \x01(\x08:\x05\x66\x61lse\x12,\n\x0f\x62oot_validation\x18\n \x03(\x0b\x32\x13.dfu.BootValidation"\x1f\n\x0cResetCommand\x12\x0f\n\x07timeout\x18\x01 \x02(\r"i\n\x07\x43ommand\x12\x1c\n\x07op_code\x18\x01 \x01(\x0e\x32\x0b.dfu.OpCode\x12\x1e\n\x04init\x18\x02 \x01(\x0b\x32\x10.dfu.InitCommand\x12 \n\x05reset\x18\x03 \x01(\x0b\x32\x11.dfu.ResetCommand"m\n\rSignedCommand\x12\x1d\n\x07\x63ommand\x18\x01 \x02(\x0b\x32\x0c.dfu.Command\x12*\n\x0esignature_type\x18\x02 \x02(\x0e\x32\x12.dfu.SignatureType\x12\x11\n\tsignature\x18\x03 \x02(\x0c"S\n\x06Packet\x12\x1d\n\x07\x63ommand\x18\x01 \x01(\x0b\x32\x0c.dfu.Command\x12*\n\x0esigned_command\x18\x02 \x01(\x0b\x32\x12.dfu.SignedCommand*\x1d\n\x06OpCode\x12\t\n\x05RESET\x10\x00\x12\x08\n\x04INIT\x10\x01*n\n\x06\x46wType\x12\x0f\n\x0b\x41PPLICATION\x10\x00\x12\x0e\n\nSOFTDEVICE\x10\x01\x12\x0e\n\nBOOTLOADER\x10\x02\x12\x19\n\x15SOFTDEVICE_BOOTLOADER\x10\x03\x12\x18\n\x14\x45XTERNAL_APPLICATION\x10\x04*D\n\x08HashType\x12\x0b\n\x07NO_HASH\x10\x00\x12\x07\n\x03\x43RC\x10\x01\x12\n\n\x06SHA128\x10\x02\x12\n\n\x06SHA256\x10\x03\x12\n\n\x06SHA512\x10\x04*t\n\x0eValidationType\x12\x11\n\rNO_VALIDATION\x10\x00\x12\x1a\n\x16VALIDATE_GENERATED_CRC\x10\x01\x12\x13\n\x0fVALIDATE_SHA256\x10\x02\x12\x1e\n\x1aVALIDATE_ECDSA_P256_SHA256\x10\x03*3\n\rSignatureType\x12\x15\n\x11\x45\x43\x44SA_P256_SHA256\x10\x00\x12\x0b\n\x07\x45\x44\x32\x35\x35\x31\x39\x10\x01'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "src.nitrokey.trussed._bootloader.nrf52_upload.dfu.dfu_cc_pb2", _globals
)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_INITCOMMAND"].fields_by_name["sd_req"]._loaded_options = None
    _globals["_INITCOMMAND"].fields_by_name["sd_req"]._serialized_options = b"\020\001"
    _globals["_OPCODE"]._serialized_start = 783
    _globals["_OPCODE"]._serialized_end = 812
    _globals["_FWTYPE"]._serialized_start = 814
    _globals["_FWTYPE"]._serialized_end = 924
    _globals["_HASHTYPE"]._serialized_start = 926
    _globals["_HASHTYPE"]._serialized_end = 994
    _globals["_VALIDATIONTYPE"]._serialized_start = 996
    _globals["_VALIDATIONTYPE"]._serialized_end = 1112
    _globals["_SIGNATURETYPE"]._serialized_start = 1114
    _globals["_SIGNATURETYPE"]._serialized_end = 1165
    _globals["_HASH"]._serialized_start = 71
    _globals["_HASH"]._serialized_end = 125
    _globals["_BOOTVALIDATION"]._serialized_start = 127
    _globals["_BOOTVALIDATION"]._serialized_end = 194
    _globals["_INITCOMMAND"]._serialized_start = 197
    _globals["_INITCOMMAND"]._serialized_end = 445
    _globals["_RESETCOMMAND"]._serialized_start = 447
    _globals["_RESETCOMMAND"]._serialized_end = 478
    _globals["_COMMAND"]._serialized_start = 480
    _globals["_COMMAND"]._serialized_end = 585
    _globals["_SIGNEDCOMMAND"]._serialized_start = 587
    _globals["_SIGNEDCOMMAND"]._serialized_end = 696
    _globals["_PACKET"]._serialized_start = 698
    _globals["_PACKET"]._serialized_end = 781
# @@protoc_insertion_point(module_scope)



import enum
from dataclasses import dataclass
from enum import Enum, IntFlag
from typing import Optional

from fido2 import cbor
from fido2.ctap import CtapError

from . import App, TimeoutException, TrussedDevice, Uuid, Version

RNG_LEN = 57
UUID_LEN = 16
VERSION_LEN = 4


@enum.unique
class AdminCommand(Enum):
    # legacy commands -- can be called directly or using the admin namespace
    UPDATE = 0x51
    REBOOT = 0x53
    RNG = 0x60
    VERSION = 0x61
    UUID = 0x62
    LOCKED = 0x63

    # new commands -- can only be called using the admin namespace
    STATUS = 0x80
    TEST_SE050 = 0x81
    GET_CONFIG = 0x82
    SET_CONFIG = 0x83
    FACTORY_RESET = 0x84
    FACTORY_RESET_APP = 0x85
    LIST_AVAILABLE_FIELDS = 0x86

    def is_legacy_command(self) -> bool:
        if self == AdminCommand.UPDATE:
            return True
        if self == AdminCommand.REBOOT:
            return True
        if self == AdminCommand.RNG:
            return True
        if self == AdminCommand.VERSION:
            return True
        if self == AdminCommand.UUID:
            return True
        if self == AdminCommand.LOCKED:
            return True
        return False


@enum.unique
class BootMode(Enum):
    FIRMWARE = enum.auto()
    BOOTROM = enum.auto()


@enum.unique
class InitStatus(IntFlag):
    NFC_ERROR = 0b0001
    INTERNAL_FLASH_ERROR = 0b0010
    EXTERNAL_FLASH_ERROR = 0b0100
    MIGRATION_ERROR = 0b1000
    SE050_ERROR = 0b00010000
    CONFIG_ERROR = 0b00100000
    RNG_ERROR = 0b01000000
    EXT_FLASH_NEED_REFORMAT = 0b10000000

    def is_error(self) -> bool:
        return self.value != 0

    def __str__(self) -> str:
        if self == 0:
            return "ok"
        errors = [error for error in InitStatus if error in self if error.name]
        value = sum(errors)
        messages = [error.name for error in errors if error.name]
        if self.value != value:
            messages.append("UNKNOWN")
        return ", ".join(messages) + " (" + hex(self.value) + ")"


@enum.unique
class Variant(Enum):
    USBIP = 0
    LPC55 = 1
    NRF52 = 2


@dataclass
class Status:
    init_status: Optional[InitStatus] = None
    ifs_blocks: Optional[int] = None
    efs_blocks: Optional[int] = None
    variant: Optional[Variant] = None


@enum.unique
class FactoryResetStatus(Enum):
    SUCCESS = 0
    NOT_CONFIRMED = 0x01
    APP_NOT_ALLOWED = 0x02
    APP_FAILED_PARSE = 0x03

    @classmethod
    def from_int(cls, i: int) -> Optional["FactoryResetStatus"]:
        for status in FactoryResetStatus:
            if status.value == i:
                return status
        return None

    @classmethod
    def check(cls, i: int, msg: str) -> None:
        status = FactoryResetStatus.from_int(i)
        if status != FactoryResetStatus.SUCCESS:
            if status is None:
                raise Exception(f"Unknown error {i:x}")
            if status == FactoryResetStatus.NOT_CONFIRMED:
                error = "Operation was not confirmed with touch"
            elif status == FactoryResetStatus.APP_NOT_ALLOWED:
                error = "The application does not support factory reset through nitropy"
            elif status == FactoryResetStatus.APP_FAILED_PARSE:
                error = "The application name must be utf-8"
            raise Exception(f"{msg}: {error}")


@enum.unique
class ConfigStatus(Enum):
    SUCCESS = 0
    READ_FAILED = 1
    WRITE_FAILED = 2
    DESERIALIZATION_FAILED = 3
    SERIALIZATION_FAILED = 4
    INVALID_KEY = 5
    INVALID_VALUE = 6
    DATA_TOO_LONG = 7

    @classmethod
    def from_int(cls, i: int) -> Optional["ConfigStatus"]:
        for status in ConfigStatus:
            if status.value == i:
                return status
        return None

    @classmethod
    def check(cls, i: int, msg: str) -> None:
        status = ConfigStatus.from_int(i)
        if status != ConfigStatus.SUCCESS:
            if status:
                error = str(status)
            else:
                error = f"unknown error {i:x}"
            raise Exception(f"{msg}: {error}")


@enum.unique
class ConfigFieldType(Enum):
    BOOL = 0
    U8 = 1

    @classmethod
    def from_int(cls, i: int) -> Optional["ConfigFieldType"]:
        for ty in ConfigFieldType:
            if ty.value == i:
                return ty
        return None

    def is_valid(self, value: str) -> bool:
        if self == ConfigFieldType.BOOL:
            return value in ["true", "false"]
        elif self == ConfigFieldType.U8:
            try:
                intval = int(value)
                if intval < 256 and intval >= 0:
                    return True
                else:
                    return False
            except ValueError:
                return False

    def __str__(self) -> str:
        if self == ConfigFieldType.BOOL:
            return "Bool"
        elif self == ConfigFieldType.U8:
            return "u8"


@dataclass
class ConfigField:
    name: str
    requires_touch_confirmation: bool
    requires_reboot: bool
    destructive: bool
    ty: ConfigFieldType


class AdminApp:
    def __init__(self, device: TrussedDevice) -> None:
        self.device = device

    def _call(
        self,
        command: AdminCommand,
        response_len: Optional[int] = None,
        data: bytes = b"",
    ) -> Optional[bytes]:
        try:
            if command.is_legacy_command():
                return self.device._call(
                    command.value,
                    command.name,
                    response_len=response_len,
                    data=data,
                )
            else:
                return self.device._call_app(
                    App.ADMIN,
                    response_len=response_len,
                    data=command.value.to_bytes(1, "big") + data,
                )
        except CtapError as e:
            if e.code == CtapError.ERR.INVALID_COMMAND:
                return None
            else:
                raise

    def is_locked(self) -> bool:
        response = self._call(AdminCommand.LOCKED, response_len=1)
        assert response is not None
        return response[0] == 1

    def reboot(self, mode: BootMode = BootMode.FIRMWARE) -> bool:
        try:
            if mode == BootMode.FIRMWARE:
                self._call(AdminCommand.REBOOT)
            elif mode == BootMode.BOOTROM:
                try:
                    self._call(AdminCommand.UPDATE)
                except CtapError as e:
                    # The admin app returns an Invalid Length error if the user confirmation
                    # request times out
                    if e.code == CtapError.ERR.INVALID_LENGTH:
                        raise TimeoutException()
                    else:
                        raise e
        except OSError as e:
            # OS error is expected as the device does not respond during the reboot
            self.device._logger.debug("ignoring OSError after reboot", exc_info=e)
        return True

    def rng(self) -> bytes:
        data = self._call(AdminCommand.RNG, response_len=RNG_LEN)
        assert data is not None
        return data

    def status(self) -> Status:
        status = Status()
        reply = self._call(AdminCommand.STATUS)
        if reply is not None:
            if not reply:
                raise ValueError("The device returned an empty status")
            status.init_status = InitStatus(reply[0])
            if len(reply) >= 4:
                status.ifs_blocks = reply[1]
                status.efs_blocks = int.from_bytes(reply[2:4], "big")
            if len(reply) >= 5:
                try:
                    status.variant = Variant(reply[4])
                except ValueError:
                    pass
        return status

    def uuid(self) -> Optional[Uuid]:
        uuid = self._call(AdminCommand.UUID)
        if uuid is None or len(uuid) == 0:
            # Firmware version 1.0.0 does not support querying the UUID
            return None
        if len(uuid) != UUID_LEN:
            raise ValueError(f"UUID response has invalid length {len(uuid)}")
        return Uuid(int.from_bytes(uuid, byteorder="big"))

    def version(self) -> Version:
        reply = self._call(AdminCommand.VERSION, data=bytes([0x01]))
        assert reply is not None
        if len(reply) == VERSION_LEN:
            version = int.from_bytes(reply, "big")
            return Version.from_int(version)
        else:
            return Version.from_str(reply.decode("utf-8"))

    def se050_tests(self) -> Optional[bytes]:
        return self._call(AdminCommand.TEST_SE050)

    def has_config(self, key: str) -> bool:
        reply = self._call(AdminCommand.GET_CONFIG, data=key.encode())
        if not reply or len(reply) < 1:
            return False
        return ConfigStatus.from_int(reply[0]) == ConfigStatus.SUCCESS

    def get_config(self, key: str) -> str:
        reply = self._call(AdminCommand.GET_CONFIG, data=key.encode())
        if not reply or len(reply) < 1:
            raise ValueError("The device returned an empty response")
        ConfigStatus.check(reply[0], "Failed to get config value")
        return reply[1:].decode()

    def set_config(self, key: str, value: str) -> None:
        request = cbor.encode({"key": key, "value": value})
        reply = self._call(AdminCommand.SET_CONFIG, data=request, response_len=1)
        assert reply
        ConfigStatus.check(reply[0], "Failed to set config value")

    def list_available_fields(self) -> list[ConfigField]:
        reply = self._call(AdminCommand.LIST_AVAILABLE_FIELDS)
        # Function not supported, return values for 1.7.0 release
        if not reply:
            return [
                ConfigField(
                    "fido.disable_skip_up_timeout",
                    False,
                    False,
                    False,
                    ConfigFieldType.BOOL,
                ),
                ConfigField(
                    "opcard.use_se050_backend",
                    True,
                    True,
                    True,
                    ConfigFieldType.BOOL,
                ),
            ]
        parsed = cbor.decode(reply)
        assert isinstance(parsed, list)
        ret = []

        for field in parsed:
            ty = ConfigFieldType.from_int(field["t"])
            if ty is None:
                ty = ConfigFieldType.BOOL

            ret.append(
                ConfigField(
                    field["n"],
                    field["c"],
                    field["r"],
                    field["d"],
                    ty,
                )
            )
        return ret

    def factory_reset(self) -> bool:
        try:
            reply = self._call(AdminCommand.FACTORY_RESET, response_len=1)
        except OSError as e:
            if e.errno == 5:
                self.device._logger.debug("ignoring OSError after reboot", exc_info=e)
                return True
            else:
                raise e
        if reply:
            FactoryResetStatus.check(reply[0], "Failed to factory reset the device")
        return reply is not None

    def factory_reset_app(self, application: str) -> bool:
        reply = self._call(
            AdminCommand.FACTORY_RESET_APP,
            data=application.encode("ascii"),
            response_len=1,
        )
        if reply:
            FactoryResetStatus.check(
                reply[0], "Failed to factory reset the application"
            )
        return reply is not None



# Copyright (c) 2016 - 2019 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
from abc import ABC, abstractmethod
from typing import Any, Callable

# Nordic Semiconductor imports

logger = logging.getLogger(__name__)


# Custom logging level for logging all transport events, including all bytes
# being transported over the wire or over the air.
# Note that this logging level is more verbose than logging.DEBUG.
TRANSPORT_LOGGING_LEVEL = 5


class DfuEvent:
    PROGRESS_EVENT = 1


class DfuTransport(ABC):
    """
    This class as an abstract base class inherited from when implementing transports.

    The class is generic in nature, the underlying implementation may have missing semantic
    than this class describes. But the intent is that the implementer shall follow the semantic as
    best as she can.
    """

    OP_CODE = {
        "CreateObject": 0x01,
        "SetPRN": 0x02,
        "CalcChecSum": 0x03,
        "Execute": 0x04,
        "ReadObject": 0x06,
        "Response": 0x60,
    }

    RES_CODE = {
        "InvalidCode": 0x00,
        "Success": 0x01,
        "NotSupported": 0x02,
        "InvalidParameter": 0x03,
        "InsufficientResources": 0x04,
        "InvalidObject": 0x05,
        "InvalidSignature": 0x06,
        "UnsupportedType": 0x07,
        "OperationNotPermitted": 0x08,
        "OperationFailed": 0x0A,
        "ExtendedError": 0x0B,
    }

    EXT_ERROR_CODE = [
        "No extended error code has been set. This error indicates an implementation problem.",
        "Invalid error code. This error code should never be used outside of development.",
        "The format of the command was incorrect. This error code is not used in the current implementation, because @ref NRF_DFU_RES_CODE_OP_CODE_NOT_SUPPORTED and @ref NRF_DFU_RES_CODE_INVALID_PARAMETER cover all possible format errors.",
        "The command was successfully parsed, but it is not supported or unknown.",
        "The init command is invalid. The init packet either has an invalid update type or it is missing required fields for the update type (for example, the init packet for a SoftDevice update is missing the SoftDevice size field).",
        "The firmware version is too low. For an application, the version must be greater than or equal to the current application. For a bootloader, it must be greater than the current version. This requirement prevents downgrade attacks."
        "",
        "The hardware version of the device does not match the required hardware version for the update.",
        "The array of supported SoftDevices for the update does not contain the FWID of the current SoftDevice.",
        "The init packet does not contain a signature, but this bootloader requires all updates to have one.",
        "The hash type that is specified by the init packet is not supported by the DFU bootloader.",
        "The hash of the firmware image cannot be calculated.",
        "The type of the signature is unknown or not supported by the DFU bootloader.",
        "The hash of the received firmware image does not match the hash in the init packet.",
        "The available space on the device is insufficient to hold the firmware.",
        "The requested firmware to update was already present on the system.",
    ]

    @abstractmethod
    def __init__(self) -> None:
        self.callbacks: dict[int, list[Callable[..., None]]] = {}

    @abstractmethod
    def open(self) -> None:
        """
        Open a port if appropriate for the transport.
        :return:
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Close a port if appropriate for the transport.
        :return:
        """
        pass

    @abstractmethod
    def send_init_packet(self, init_packet: bytes) -> None:
        """
        Send init_packet to device.

        This call will block until init_packet is sent and transfer of packet is complete.

        :param init_packet: Init packet as a str.
        :return:
        """
        pass

    @abstractmethod
    def send_firmware(self, firmware: bytes) -> None:
        """
        Start sending firmware to device.

        This call will block until transfer of firmware is complete.

        :param firmware:
        :return:
        """
        pass

    def register_events_callback(
        self, event_type: int, callback: Callable[..., None]
    ) -> None:
        """
        Register a callback.

        :param DfuEvent callback:
        :return: None
        """
        if event_type not in self.callbacks:
            self.callbacks[event_type] = []

        self.callbacks[event_type].append(callback)

    def _send_event(self, event_type: int, **kwargs: Any) -> None:
        """
        Method for sending events to registered callbacks.

        If callbacks throws exceptions event propagation will stop and this method be part of the track trace.

        :param DfuEvent event_type:
        :param kwargs: Arguments to callback function
        :return:
        """
        if event_type in list(self.callbacks.keys()):
            for callback in self.callbacks[event_type]:
                callback(**kwargs)



# Copyright 2022 Nitrokey Developers
#
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.

import enum
import importlib.metadata
import logging
import platform
import re
import time
from abc import ABC, abstractmethod
from collections.abc import Set
from contextlib import contextmanager
from importlib.metadata import PackageNotFoundError
from io import BytesIO
from typing import TYPE_CHECKING, Any, Callable, Iterator, List, Optional, Tuple, Union

from nitrokey._helpers import Retries
from nitrokey.trussed import TimeoutException, TrussedBase, Version
from nitrokey.trussed._base import Model
from nitrokey.trussed._bootloader import (
    FirmwareContainer,
    TrussedBootloader,
    Variant,
    get_model_data,
    validate_firmware_image,
)
from nitrokey.trussed._bootloader.lpc55_upload.mboot.exceptions import (
    McuBootConnectionError,
)
from nitrokey.trussed._device import TrussedDevice
from nitrokey.trussed.admin_app import BootMode, Status
from nitrokey.trussed.admin_app import Variant as AdminAppVariant
from nitrokey.updates import Asset, Release, Repository

if TYPE_CHECKING:
    import typing_extensions

logger = logging.getLogger(__name__)


@enum.unique
class Warning(enum.Enum):
    """
    A warning that can occur during a firmware update.

    By default, these warnings abort the firmware update.  This enum can be used to select types
    of warnings that should be ignored and not cause the firmware update to fail.
    """

    IFS_MIGRATION_V2 = "ifs-migration-v2"
    MISSING_STATUS = "missing-status"
    SDK_VERSION = "sdk-version"
    UPDATE_FROM_BOOTLOADER = "update-from-bootloader"

    @property
    def message(self) -> str:
        if self == Warning.IFS_MIGRATION_V2:
            return (
                "Not enough space on the internal filesystem to perform the firmware"
                " update. See the release notes for more information:"
                " https://github.com/Nitrokey/nitrokey-3-firmware/releases/tag/v1.8.2"
            )
        if self == Warning.MISSING_STATUS:
            return (
                "Could not determine the device state as the current firmware is too old."
                " Please update to firmware version v1.3.1 first."
            )
        if self == Warning.SDK_VERSION:
            return (
                "Your Nitrokey SDK version is outdated.  Please update this program to the latest"
                " version and try again."
            )
        if self == Warning.UPDATE_FROM_BOOTLOADER:
            return (
                "The current state of the device cannot be checked as it is already in bootloader"
                " mode. Please review the release notes at:"
                " https://github.com/Nitrokey/nitrokey-3-firmware/releases"
            )

        if TYPE_CHECKING:
            typing_extensions.assert_never(self)

        return self.value

    @classmethod
    def from_str(cls, s: str) -> "Warning":
        for w in cls:
            if w.value == s:
                return w
        raise ValueError(f"Unexpected update warning id: {s}")


@enum.unique
class _Migration(enum.Enum):
    # IFS migration to use journaling on the NRF52 introduced in v1.3.0 (NK3)
    NRF_IFS_MIGRATION = enum.auto()
    # IFS migration to filesystem layout 2 (FIDO2 RK migration) in v1.8.2 (NK3)
    IFS_MIGRATION_V2 = enum.auto()

    @classmethod
    def get(
        cls,
        model: Model,
        variant: Union[Variant, AdminAppVariant],
        current: Optional[Version],
        new: Version,
    ) -> frozenset["_Migration"]:
        if model != Model.NK3:
            return frozenset()

        if isinstance(variant, AdminAppVariant):
            if variant == AdminAppVariant.USBIP:
                raise ValueError("Cannot perform firmware update for USBIP runner")
            elif variant == AdminAppVariant.LPC55:
                variant = Variant.LPC55
            elif variant == AdminAppVariant.NRF52:
                variant = Variant.NRF52
            else:
                if TYPE_CHECKING:
                    typing_extensions.assert_never(variant)

                raise ValueError(f"Unsupported device variant: {variant}")

        migrations = set()

        if variant == Variant.NRF52:
            if (
                current is None
                or current <= Version(1, 2, 2)
                and new >= Version(1, 3, 0)
            ):
                migrations.add(cls.NRF_IFS_MIGRATION)

        ifs_migration_v2 = Version(1, 8, 2)
        if (
            current is not None
            and current < ifs_migration_v2
            and new >= ifs_migration_v2
        ):
            migrations.add(cls.IFS_MIGRATION_V2)

        return frozenset(migrations)


def get_firmware_repository(model: Model) -> Repository:
    data = get_model_data(model)
    return Repository(owner="Nitrokey", name=data.firmware_repository_name)


def get_firmware_update(model: Model, release: Release) -> Asset:
    data = get_model_data(model)
    return release.require_asset(re.compile(data.firmware_pattern_string))


def _get_extra_information(migrations: Set[_Migration]) -> List[str]:
    """Return additional information for the device after update based on update-path"""

    out = []
    if _Migration.NRF_IFS_MIGRATION in migrations:
        out += [
            "",
            "During this update process the internal filesystem will be migrated!",
            "- Migration will only work, if your internal filesystem does not contain more than 45 Resident Keys. If you have more please remove some.",
            "- After the update it might take up to 3 minutes for the first boot.",
            "Never unplug the device while the LED is active!",
        ]
    return out


def _get_finalization_wait_retries(migrations: Set[_Migration]) -> int:
    """Return number of retries to wait for the device after update based on update-path"""

    out = 60
    if _Migration.NRF_IFS_MIGRATION in migrations:
        # max time 150secs == 300 retries
        out = 500
    return out


class UpdateUi(ABC):
    @abstractmethod
    def error(self, *msgs: Any) -> Exception:
        pass

    @abstractmethod
    def show_warning(self, warning: Warning) -> None:
        pass

    @abstractmethod
    def raise_warning(self, warning: Warning) -> Exception:
        pass

    @abstractmethod
    def abort(self, *msgs: Any) -> Exception:
        pass

    @abstractmethod
    def abort_downgrade(self, current: Version, image: Version) -> Exception:
        pass

    @abstractmethod
    def abort_pynitrokey_version(
        self, current: Version, required: Version
    ) -> Exception:
        pass

    @abstractmethod
    def confirm_download(self, current: Optional[Version], new: Version) -> None:
        pass

    @abstractmethod
    def confirm_update(self, current: Optional[Version], new: Version) -> None:
        pass

    @abstractmethod
    def confirm_pynitrokey_version(self, current: Version, required: Version) -> None:
        pass

    @abstractmethod
    def confirm_extra_information(self, extra_info: List[str]) -> None:
        pass

    @abstractmethod
    def confirm_update_same_version(self, version: Version) -> None:
        pass

    @abstractmethod
    def pre_bootloader_hint(self) -> None:
        pass

    @abstractmethod
    def request_bootloader_confirmation(self) -> None:
        pass

    @abstractmethod
    @contextmanager
    def download_progress_bar(self, desc: str) -> Iterator[Callable[[int, int], None]]:
        pass

    @abstractmethod
    @contextmanager
    def update_progress_bar(self) -> Iterator[Callable[[int, int], None]]:
        pass

    @abstractmethod
    @contextmanager
    def finalization_progress_bar(self) -> Iterator[Callable[[int, int], None]]:
        pass


class DeviceHandler(ABC):
    @abstractmethod
    def await_bootloader(self, model: Model) -> TrussedBootloader: ...

    @abstractmethod
    def await_device(
        self,
        model: Model,
        wait_retries: Optional[int],
        callback: Optional[Callable[[int, int], None]],
    ) -> TrussedDevice: ...


class Updater:
    def __init__(
        self,
        ui: UpdateUi,
        device_handler: DeviceHandler,
        ignore_warnings: Set[Warning] = frozenset(),
    ) -> None:
        self.ui = ui
        self.device_handler = device_handler
        self.ignore_warnings = ignore_warnings

    def _trigger_warning(self, warning: Warning) -> None:
        if warning in self.ignore_warnings:
            self.ui.show_warning(warning)
        else:
            raise self.ui.raise_warning(warning)

    def update(
        self,
        device: TrussedBase,
        image: Optional[str] = None,
        update_version: Optional[str] = None,
        ignore_pynitrokey_version: bool = False,
    ) -> Tuple[Version, Status]:
        model = device.model

        update_from_bootloader = False
        current_version = None
        status = None
        if isinstance(device, TrussedBootloader):
            update_from_bootloader = True
            self._trigger_warning(Warning.UPDATE_FROM_BOOTLOADER)
        elif isinstance(device, TrussedDevice):
            current_version = device.admin.version()
            status = device.admin.status()
        else:
            raise self.ui.error(f"Unexpected Trussed device: {device}")

        logger.info(f"Firmware version before update: {current_version or ''}")
        container = self._prepare_update(model, image, update_version, current_version)

        if not update_from_bootloader:
            if status is None:
                if model == Model.NK3:
                    if container.version > Version.from_str("1.3.1"):
                        self._trigger_warning(Warning.MISSING_STATUS)
                else:
                    self.ui.error(f"Missing status for {model} device")

        self._check_minimum_version(container, ignore_pynitrokey_version)

        self.ui.confirm_update(current_version, container.version)

        migrations = None
        if status is not None and status.variant is not None:
            migrations = self._check_migrations(
                model, status.variant, current_version, container.version, status
            )
        elif isinstance(device, TrussedBootloader):
            migrations = self._check_migrations(
                model, device.variant, current_version, container.version, status
            )

        with self._get_bootloader(device) as bootloader:
            if bootloader.variant not in container.images:
                raise self.ui.error(
                    "The firmware release does not contain an image for the "
                    f"{bootloader.variant.value} hardware variant"
                )
            try:
                validate_firmware_image(
                    bootloader.variant,
                    container.images[bootloader.variant],
                    container.version,
                    model,
                )
            except Exception as e:
                raise self.ui.error("Failed to validate firmware image", e)

            if migrations is None:
                migrations = self._check_migrations(
                    model,
                    bootloader.variant,
                    current_version,
                    container.version,
                    status,
                )

            self._perform_update(bootloader, container)

        wait_retries = _get_finalization_wait_retries(migrations)
        with self.ui.finalization_progress_bar() as callback:
            with self.device_handler.await_device(
                model, wait_retries, callback
            ) as device:
                version = device.admin.version()
                if version != container.version:
                    raise self.ui.error(
                        f"The firmware update to {container.version} was successful, but the "
                        f"firmware is still reporting version {version}."
                    )
                status = device.admin.status()

        return container.version, status

    def _prepare_update(
        self,
        model: Model,
        image: Optional[str],
        version: Optional[str],
        current_version: Optional[Version],
    ) -> FirmwareContainer:
        if image:
            try:
                container = FirmwareContainer.parse(image, model)
            except Exception as e:
                raise self.ui.error("Failed to parse firmware container", e)
            self._validate_version(current_version, container.version)
            return container
        else:
            repository = get_firmware_repository(model)
            if version:
                try:
                    logger.info(f"Downloading firmare version {version}")
                    release = repository.get_release(version)
                except Exception as e:
                    raise self.ui.error(f"Failed to get firmware release {version}", e)
            else:
                try:
                    release = repository.get_latest_release()
                    logger.info(f"Latest firmware version: {release}")
                except Exception as e:
                    raise self.ui.error("Failed to find latest firmware release", e)

            try:
                release_version = Version.from_v_str(release.tag)
            except ValueError as e:
                raise self.ui.error("Failed to parse version from release tag", e)
            self._validate_version(current_version, release_version)
            self.ui.confirm_download(current_version, release_version)
            return self._download_update(model, release)

    def _download_update(self, model: Model, release: Release) -> FirmwareContainer:
        try:
            update = get_firmware_update(model, release)
        except Exception as e:
            raise self.ui.error(
                f"Failed to find firmware image for release {release}",
                e,
            )

        try:
            logger.info(f"Trying to download firmware update from URL: {update.url}")

            with self.ui.download_progress_bar(update.tag) as callback:
                data = update.read(callback=callback)
        except Exception as e:
            raise self.ui.error(
                f"Failed to download latest firmware update {update.tag}", e
            )

        try:
            container = FirmwareContainer.parse(BytesIO(data), model)
        except Exception as e:
            raise self.ui.error(
                f"Failed to parse firmware container for {update.tag}", e
            )

        release_version = Version.from_v_str(release.tag)
        if release_version != container.version:
            raise self.ui.error(
                f"The firmware container for {update.tag} has the version {container.version}"
            )

        return container

    def _check_minimum_version(
        self, container: FirmwareContainer, ignore_pynitrokey_version: bool
    ) -> None:
        if container.sdk:
            try:
                sdk_version = Version.from_str(importlib.metadata.version("nitrokey"))
            except PackageNotFoundError:
                raise self.ui.error("Failed to determine the Nitrokey SDK version")

            if container.sdk > sdk_version:
                logger.warning(
                    f"Minimum SDK version required for update is {container.sdk} (current version: {sdk_version})"
                )
                self._trigger_warning(Warning.SDK_VERSION)
        elif container.pynitrokey:
            # The minimum pynitrokey version has been replaced by the minimum SDK version, so we
            # only check it if there is no minimum SDK version set.

            # this is the version of pynitrokey when we moved to the SDK
            pynitrokey_version = Version.from_str("0.4.49")
            if container.pynitrokey > pynitrokey_version:
                if ignore_pynitrokey_version:
                    self.ui.confirm_pynitrokey_version(
                        current=pynitrokey_version, required=container.pynitrokey
                    )
                else:
                    raise self.ui.abort_pynitrokey_version(
                        current=pynitrokey_version, required=container.pynitrokey
                    )

    def _validate_version(
        self,
        current_version: Optional[Version],
        new_version: Version,
    ) -> None:
        logger.info(f"Current firmware version: {current_version}")
        logger.info(f"Updated firmware version: {new_version}")

        if current_version:
            if current_version.core() > new_version.core():
                raise self.ui.abort_downgrade(current_version, new_version)
            elif current_version == new_version:
                if current_version.complete and new_version.complete:
                    same_version = current_version
                else:
                    same_version = current_version.core()
                self.ui.confirm_update_same_version(same_version)

    @contextmanager
    def _get_bootloader(self, device: TrussedBase) -> Iterator[TrussedBootloader]:
        model = device.model
        if isinstance(device, TrussedDevice):
            self.ui.request_bootloader_confirmation()
            try:
                device.admin.reboot(BootMode.BOOTROM)
            except TimeoutException:
                raise self.ui.abort(
                    "The reboot was not confirmed with the touch button"
                )

            # needed for udev to properly handle new device
            time.sleep(1)

            self.ui.pre_bootloader_hint()

            exc = None
            for t in Retries(3):
                logger.debug(f"Trying to connect to bootloader ({t})")
                try:
                    with self.device_handler.await_bootloader(model) as bootloader:
                        # noop to test communication
                        bootloader.uuid
                        yield bootloader
                        break
                except McuBootConnectionError as e:
                    logger.debug("Received connection error", exc_info=True)
                    exc = e
            else:
                msgs = [f"Failed to connect to {model} bootloader"]
                if platform.system() == "Linux":
                    msgs += ["Are the Nitrokey udev rules installed and active?"]
                raise self.ui.error(*msgs, exc)
        elif isinstance(device, TrussedBootloader):
            yield device
        else:
            raise self.ui.error(f"Unexpected {model} device: {device}")

    def _check_migrations(
        self,
        model: Model,
        variant: Union[Variant, AdminAppVariant],
        current_version: Optional[Version],
        new_version: Version,
        status: Optional[Status],
    ) -> frozenset["_Migration"]:
        try:
            migrations = _Migration.get(
                model=model,
                variant=variant,
                current=current_version,
                new=new_version,
            )
        except ValueError as e:
            raise self.ui.error(str(e))

        txt = _get_extra_information(migrations)
        if len(txt) > 0:
            self.ui.confirm_extra_information(txt)

        if _Migration.IFS_MIGRATION_V2 in migrations:
            if status and status.ifs_blocks is not None and status.ifs_blocks < 5:
                self._trigger_warning(Warning.IFS_MIGRATION_V2)

        return migrations

    def _perform_update(
        self, device: TrussedBootloader, container: FirmwareContainer
    ) -> None:
        logger.debug("Starting firmware update")
        image = container.images[device.variant]
        with self.ui.update_progress_bar() as callback:
            try:
                device.update(image, callback=callback)
            except Exception as e:
                raise self.ui.error("Failed to perform firmware update", e)
        logger.debug("Firmware update finished successfully")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2024 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""OpenSSL implementation Hash algorithms."""

# Used security modules

from math import ceil

from cryptography.hazmat.primitives import hashes

from ..exceptions import SPSDKError
from ..utils.misc import Endianness
from ..utils.spsdk_enum import SpsdkEnum


class EnumHashAlgorithm(SpsdkEnum):
    """Hash algorithm enum."""

    SHA1 = (0, "sha1", "SHA1")
    SHA256 = (1, "sha256", "SHA256")
    SHA384 = (2, "sha384", "SHA384")
    SHA512 = (3, "sha512", "SHA512")
    MD5 = (4, "md5", "MD5")
    SM3 = (5, "sm3", "SM3")


def get_hash_algorithm(algorithm: EnumHashAlgorithm) -> hashes.HashAlgorithm:
    """For specified name return hashes algorithm instance.

    :param algorithm: Algorithm type enum
    :return: instance of algorithm class
    :raises SPSDKError: If algorithm not found
    """
    algo_cls = getattr(
        hashes, algorithm.label.upper(), None
    )  # hack: get class object by name
    if algo_cls is None:
        raise SPSDKError(f"Unsupported algorithm: hashes.{algorithm.label.upper()}")

    algo = algo_cls()
    assert isinstance(algo, hashes.HashAlgorithm)
    return algo


def get_hash_length(algorithm: EnumHashAlgorithm) -> int:
    """For specified name return hash binary length.

    :param algorithm: Algorithm type enum
    :return: Hash length
    :raises SPSDKError: If algorithm not found
    """
    return get_hash_algorithm(algorithm).digest_size


class Hash:
    """SPSDK Hash Class."""

    def __init__(self, algorithm: EnumHashAlgorithm = EnumHashAlgorithm.SHA256) -> None:
        """Initialize hash object.

        :param algorithm: Algorithm type enum, defaults to EnumHashAlgorithm.SHA256
        """
        self.hash_obj = hashes.Hash(get_hash_algorithm(algorithm))

    def update(self, data: bytes) -> None:
        """Update the hash by new data.

        :param data: Data to be hashed
        """
        self.hash_obj.update(data)

    def update_int(self, value: int) -> None:
        """Update the hash by new integer value as is.

        :param value: Integer value to be hashed
        """
        data = value.to_bytes(
            length=ceil(value.bit_length() / 8), byteorder=Endianness.BIG.value
        )
        self.update(data)

    def finalize(self) -> bytes:
        """Finalize the hash and return the hash value.

        :returns: Computed hash
        """
        return self.hash_obj.finalize()


def get_hash(
    data: bytes, algorithm: EnumHashAlgorithm = EnumHashAlgorithm.SHA256
) -> bytes:
    """Return a HASH from input data with specified algorithm.

    :param data: Input data in bytes
    :param algorithm: Algorithm type enum
    :return: Hash-ed bytes
    :raises SPSDKError: If algorithm not found
    """
    hash_obj = hashes.Hash(get_hash_algorithm(algorithm))
    hash_obj.update(data)
    return hash_obj.finalize()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Generic commands implementation."""
from abc import ABC, abstractmethod


class CmdResponseBase(ABC):
    """Response base format class."""

    @abstractmethod
    def __str__(self) -> str:
        """Get object info."""

    @property
    @abstractmethod
    def value(self) -> int:
        """Return a integer representation of the response."""


class CmdPacketBase(ABC):
    """COmmand protocol base."""

    @abstractmethod
    def to_bytes(self, padding: bool = True) -> bytes:
        """Serialize CmdPacket into bytes.

        :param padding: If True, add padding to specific size
        :return: Serialized object into bytes
        """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""MBoot protocol base."""
from ...utils.interfaces.protocol.protocol_base import ProtocolBase


class MbootProtocolBase(ProtocolBase):
    """MBoot protocol base class."""

    allow_abort: bool = False
    need_data_split: bool = True



# Copyright 2021-2024 Nitrokey Developers
#
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.

import enum
import logging
import platform
import sys
from abc import abstractmethod
from enum import Enum
from typing import List, Optional, Sequence, TypeVar, Union

from fido2.hid import CtapHidDevice, list_descriptors, open_device

from ._base import TrussedBase
from ._utils import Fido2Certs, Uuid

T = TypeVar("T", bound="TrussedDevice")

logger = logging.getLogger(__name__)


@enum.unique
class App(Enum):
    """Vendor-specific CTAPHID commands for Trussed apps."""

    SECRETS = 0x70
    PROVISIONER = 0x71
    ADMIN = 0x72


class TrussedDevice(TrussedBase):
    def __init__(
        self, device: CtapHidDevice, fido2_certs: Sequence[Fido2Certs]
    ) -> None:
        self._validate_vid_pid(device.descriptor.vid, device.descriptor.pid)

        self.device = device
        self.fido2_certs = fido2_certs
        self._path = _device_path_to_str(device.descriptor.path)
        self._logger = logger.getChild(self._path)

        from .admin_app import AdminApp

        self.admin = AdminApp(self)
        self.admin.status()

    @property
    def path(self) -> str:
        return self._path

    def close(self) -> None:
        self.device.close()

    def reboot(self) -> bool:
        from .admin_app import BootMode

        return self.admin.reboot(BootMode.FIRMWARE)

    def uuid(self) -> Optional[Uuid]:
        return self.admin.uuid()

    def wink(self) -> None:
        self.device.wink()

    def _call(
        self,
        command: int,
        command_name: str,
        response_len: Optional[int] = None,
        data: bytes = b"",
    ) -> bytes:
        response = self.device.call(command, data=data)
        if response_len is not None and response_len != len(response):
            raise ValueError(
                f"The response for the CTAPHID {command_name} command has an unexpected length "
                f"(expected: {response_len}, actual: {len(response)})"
            )
        return response

    def _call_app(
        self,
        app: App,
        response_len: Optional[int] = None,
        data: bytes = b"",
    ) -> bytes:
        return self._call(app.value, app.name, response_len, data)

    @classmethod
    @abstractmethod
    def from_device(cls: type[T], device: CtapHidDevice) -> T: ...

    @classmethod
    def open(cls: type[T], path: str) -> Optional[T]:
        try:
            if platform.system() == "Windows":
                device = open_device(bytes(path, "utf-8"))
            else:
                device = open_device(path)
        except Exception:
            logger.warn(f"No CTAPHID device at path {path}", exc_info=sys.exc_info())
            return None
        try:
            return cls.from_device(device)
        except ValueError:
            logger.warn(f"No Nitrokey device at path {path}", exc_info=sys.exc_info())
            return None

    @classmethod
    @abstractmethod
    def list(cls: type[T]) -> List[T]: ...

    @classmethod
    def _list_vid_pid(cls: type[T], vid: int, pid: int) -> List[T]:
        descriptors = [
            desc
            for desc in list_descriptors()  # type: ignore
            if desc.vid == vid and desc.pid == pid
        ]
        return [cls.from_device(open_device(desc.path)) for desc in descriptors]


def _device_path_to_str(path: Union[bytes, str]) -> str:
    """
    Converts a device path as returned by the fido2 library to a string.

    Typically, the path already is a string.  Only on Windows, a bytes object
    using an ANSI encoding is used instead.  We use the ISO 8859-1 encoding to
    decode the string which should work for all systems.
    """
    if isinstance(path, bytes):
        return path.decode("iso-8859-1", errors="ignore")
    else:
        return path



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2016-2018 Martin Olejar
# Copyright 2019-2024 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""USB Mboot interface implementation."""

from ...mboot.protocol.bulk_protocol import MbootBulkProtocol
from ...utils.interfaces.device.usb_device import UsbDevice

USB_DEVICES = {
    # NAME   | VID   | PID
    "MKL27": (0x15A2, 0x0073),
    "LPC55": (0x1FC9, 0x0021),
    "IMXRT": (0x1FC9, 0x0135),
    "MXRT10": (0x15A2, 0x0073),  # this is ID of flash-loader for RT101x
    "MXRT20": (0x15A2, 0x0073),  # this is ID of flash-loader for RT102x
    "MXRT50": (0x15A2, 0x0073),  # this is ID of flash-loader for RT105x
    "MXRT60": (0x15A2, 0x0073),  # this is ID of flash-loader for RT106x
    "LPC55xx": (0x1FC9, 0x0020),
    "LPC551x": (0x1FC9, 0x0022),
    "RT6xx": (0x1FC9, 0x0021),
    "RT5xx_A": (0x1FC9, 0x0020),
    "RT5xx_B": (0x1FC9, 0x0023),
    "RT5xx_C": (0x1FC9, 0x0023),
    "RT5xx": (0x1FC9, 0x0023),
    "RT6xxM": (0x1FC9, 0x0024),
    "LPC553x": (0x1FC9, 0x0025),
    "MCXN9xx": (0x1FC9, 0x014F),
    "MCXA1xx": (0x1FC9, 0x0155),
    "MCXN23x": (0x1FC9, 0x0158),
}


class MbootUSBInterface(MbootBulkProtocol):
    """USB interface."""

    identifier = "usb"
    device: UsbDevice
    usb_devices = USB_DEVICES

    def __init__(self, device: UsbDevice) -> None:
        """Initialize the MbootUSBInterface object.

        :param device: The device instance
        """
        assert isinstance(device, UsbDevice)
        super().__init__(device=device)

    @property
    def name(self) -> str:
        """Get the name of the device."""
        assert isinstance(self.device, UsbDevice)
        for name, value in self.usb_devices.items():
            if value[0] == self.device.vid and value[1] == self.device.pid:
                return name
        return "Unknown"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2023-2024 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Protocol base."""
from abc import ABC, abstractmethod
from types import TracebackType
from typing import TYPE_CHECKING, Optional, Type, Union

from ....utils.interfaces.commands import CmdPacketBase, CmdResponseBase
from ....utils.interfaces.device.base import DeviceBase

if TYPE_CHECKING:
    from typing_extensions import Self


class ProtocolBase(ABC):
    """Protocol base class."""

    device: DeviceBase
    identifier: str

    def __init__(self, device: DeviceBase) -> None:
        """Initialize the MbootSerialProtocol object.

        :param device: The device instance
        """
        self.device = device

    def __str__(self) -> str:
        return f"identifier='{self.identifier}', device={self.device}"

    def __enter__(self) -> "Self":
        self.open()
        return self

    def __exit__(
        self,
        exception_type: Optional[Type[Exception]] = None,
        exception_value: Optional[Exception] = None,
        traceback: Optional[TracebackType] = None,
    ) -> None:
        self.close()

    @abstractmethod
    def open(self) -> None:
        """Open the interface."""

    @abstractmethod
    def close(self) -> None:
        """Close the interface."""

    @property
    @abstractmethod
    def is_opened(self) -> bool:
        """Indicates whether interface is open."""

    @abstractmethod
    def write_command(self, packet: CmdPacketBase) -> None:
        """Write command to the device.

        :param packet: Command packet to be sent
        """

    @abstractmethod
    def write_data(self, data: bytes) -> None:
        """Write data to the device.

        :param data: Data to be send
        """

    @abstractmethod
    def read(self, length: Optional[int] = None) -> Union[CmdResponseBase, bytes]:
        """Read data from device.

        :return: read data
        """



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2024 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""The module provides support for Root Key Hash table."""

from abc import abstractmethod
from typing import TYPE_CHECKING, List

from ...crypto.hash import EnumHashAlgorithm, get_hash
from ...exceptions import SPSDKError

if TYPE_CHECKING:
    from typing_extensions import Self


class RKHT:
    """Root Key Hash Table class."""

    def __init__(self, rkh_list: List[bytes]) -> None:
        """Initialization of Root Key Hash Table class.

        :param rkh_list: List of Root Key Hashes
        """
        if len(rkh_list) > 4:
            raise SPSDKError("Number of Root Key Hashes can not be larger than 4.")
        self.rkh_list = rkh_list

    @abstractmethod
    def rkth(self) -> bytes:
        """Root Key Table Hash.

        :return: Hash of hashes of public keys.
        """

    @property
    def hash_algorithm(self) -> EnumHashAlgorithm:
        """Used hash algorithm name."""
        if not len(self.rkh_list) > 0:
            raise SPSDKError("Unknown hash algorighm name. No root key hashes.")
        return EnumHashAlgorithm.from_label(f"sha{self.hash_algorithm_size}")

    @property
    def hash_algorithm_size(self) -> int:
        """Used hash algorithm size in bites."""
        if not len(self.rkh_list) > 0:
            raise SPSDKError("Unknown hash algorithm size. No public keys provided.")
        return len(self.rkh_list[0]) * 8


class RKHTv1(RKHT):
    """Root Key Hash Table class for cert block v1."""

    RKHT_SIZE = 4
    RKH_SIZE = 32

    def __init__(
        self,
        rkh_list: List[bytes],
    ) -> None:
        """Initialization of Root Key Hash Table class.

        :param rkh_list: List of Root Key Hashes
        """
        for key_hash in rkh_list:
            if len(key_hash) != self.RKH_SIZE:
                raise SPSDKError(f"Invalid key hash size: {len(key_hash)}")
        super().__init__(rkh_list)

    @property
    def hash_algorithm(self) -> EnumHashAlgorithm:
        """Used Hash algorithm name."""
        return EnumHashAlgorithm.SHA256

    def export(self) -> bytes:
        """Export RKHT as bytes."""
        rotk_table = b""
        for i in range(self.RKHT_SIZE):
            if i < len(self.rkh_list) and self.rkh_list[i]:
                rotk_table += self.rkh_list[i]
            else:
                rotk_table += bytes(self.RKH_SIZE)
        if len(rotk_table) != self.RKH_SIZE * self.RKHT_SIZE:
            raise SPSDKError("Invalid length of data.")
        return rotk_table

    @classmethod
    def parse(cls, rkht: bytes) -> "Self":
        """Parse Root Key Hash Table into RKHTv1 object.

        :param rkht: Valid RKHT table
        """
        rotkh_len = len(rkht) // cls.RKHT_SIZE
        offset = 0
        key_hashes = []
        for _ in range(cls.RKHT_SIZE):
            key_hashes.append(rkht[offset : offset + rotkh_len])
            offset += rotkh_len
        return cls(key_hashes)

    def rkth(self) -> bytes:
        """Root Key Table Hash.

        :return: Hash of Hashes of public key.
        """
        rotkh = get_hash(self.export(), self.hash_algorithm)
        return rotkh



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Low level Hid device."""
import logging
import warnings
from typing import TYPE_CHECKING, List, Optional

from ....exceptions import SPSDKConnectionError, SPSDKError
from ....utils.exceptions import SPSDKTimeoutError
from ....utils.interfaces.device.base import DeviceBase

if TYPE_CHECKING:
    import hid

logger = logging.getLogger(__name__)


class UsbDevice(DeviceBase):
    """USB device class."""

    def __init__(
        self,
        vid: int,
        pid: int,
        path: bytes,
        vendor_name: str,
        product_name: str,
        interface_number: int,
    ) -> None:
        """Initialize the USB interface object."""
        self._opened = False
        self.vid = vid
        self.pid = pid
        self.path = path
        self.vendor_name = vendor_name
        self.product_name = product_name
        self.interface_number = interface_number
        self._timeout = 2000
        self._device: Optional["hid.device"] = None

    @property
    def timeout(self) -> int:
        """Timeout property."""
        return self._timeout

    @timeout.setter
    def timeout(self, value: int) -> None:
        """Timeout property setter."""
        self._timeout = value

    @property
    def is_opened(self) -> bool:
        """Indicates whether device is open.

        :return: True if device is open, False othervise.
        """
        return self._device is not None

    def open(self) -> None:
        """Open the interface.

        :raises SPSDKError: if device is already opened
        :raises SPSDKConnectionError: if the device can not be opened
        """
        import hid

        logger.debug(f"Opening the Interface: {str(self)}")
        if self.is_opened:
            # This would get HID_DEVICE into broken state
            raise SPSDKError("Can't open already opened device")
        try:
            self._device = hid.device()
            self._device.open_path(self.path)
        except Exception as error:
            self._device = None
            raise SPSDKConnectionError(
                f"Unable to open device '{str(self)}'"
            ) from error

    def close(self) -> None:
        """Close the interface.

        :raises SPSDKConnectionError: if no device is available
        :raises SPSDKConnectionError: if the device can not be opened
        """
        logger.debug(f"Closing the Interface: {str(self)}")
        if self._device is not None:
            try:
                self._device.close()
                self._device = None
            except Exception as error:
                raise SPSDKConnectionError(
                    f"Unable to close device '{str(self)}'"
                ) from error

    def read(self, length: int, timeout: Optional[int] = None) -> bytes:
        """Read data on the IN endpoint associated to the HID interface.

        :return: Return CmdResponse object.
        :raises SPSDKConnectionError: Raises an error if device is not opened for reading
        :raises SPSDKConnectionError: Raises if device is not available
        :raises SPSDKConnectionError: Raises if reading fails
        :raises SPSDKTimeoutError: Time-out
        """
        timeout = timeout or self.timeout
        if self._device is None:
            raise SPSDKConnectionError("Device is not opened for reading")
        try:
            data = self._device.read(length, timeout_ms=timeout)
        except Exception as e:
            raise SPSDKConnectionError(str(e)) from e
        if not data:
            logger.error("Cannot read from HID device")
            raise SPSDKTimeoutError()
        return bytes(data)

    def write(self, data: bytes, timeout: Optional[int] = None) -> None:
        """Send data to device.

        :param data: Data to send
        :param timeout: Timeout to be used
        :raises SPSDKConnectionError: Sending data to device failure
        """
        timeout = timeout or self.timeout
        if self._device is None:
            raise SPSDKConnectionError("Device is not opened for writing")
        try:
            bytes_written = self._device.write(data)
        except Exception as e:
            raise SPSDKConnectionError(str(e)) from e
        if bytes_written < 0 or bytes_written < len(data):
            raise SPSDKConnectionError(
                f"Invalid size of written bytes has been detected: {bytes_written} != {len(data)}"
            )

    def __str__(self) -> str:
        """Return information about the USB interface."""
        return (
            f"{self.product_name:s} (0x{self.vid:04X}, 0x{self.pid:04X})"
            f"path={self.path!r}"
        )

    def __hash__(self) -> int:
        return hash(self.path)

    @classmethod
    def enumerate(
        cls,
        vid: Optional[int] = None,
        pid: Optional[int] = None,
        path: Optional[str] = None,
    ) -> List["UsbDevice"]:
        """Get list of all connected devices which matches device_id."""
        try:
            import hid
        except ImportError as err:
            logger.warning("Failed to import hid module", exc_info=True)
            warnings.warn(
                f"Failed to list LPC55 bootloaders due to a missing library: {err}",
                category=RuntimeWarning,
            )
            return []

        devices = []

        # iterate on all devices found
        for dev in hid.enumerate(vendor_id=vid or 0, product_id=pid or 0):
            if path is None or dev["path"] == path.encode():
                new_device = cls(
                    vid=dev["vendor_id"],
                    pid=dev["product_id"],
                    path=dev["path"],
                    vendor_name=dev["manufacturer_string"],
                    product_name=dev["product_string"],
                    interface_number=dev["interface_number"],
                )
                devices.append(new_device)
        return devices



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2016-2018 Martin Olejar
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module implementing communication with the MCU Bootloader."""
#



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""OpenSSL implementation for HMAC packet authentication."""

# Used security modules
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import hmac as hmac_cls


def hmac_sha256(key: bytes, data: bytes) -> bytes:
    """Return a HMAC from data with specified key and algorithm.

    :param key: The key in bytes format
    :param data: Input data in bytes format
    :return: HMAC bytes
    """
    hmac_obj = hmac_cls.HMAC(key, hashes.SHA256())
    hmac_obj.update(data)
    return hmac_obj.finalize()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Custom enum extension."""
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, List, Optional, Union

if TYPE_CHECKING:
    from typing_extensions import Self

from ..exceptions import SPSDKKeyError, SPSDKTypeError


@dataclass(frozen=True)
class SpsdkEnumMember:
    """SPSDK Enum member."""

    tag: int
    label: str
    description: Optional[str] = None


class SpsdkEnum(SpsdkEnumMember, Enum):
    """SPSDK Enum type."""

    def __eq__(self, __value: object) -> bool:
        return self.tag == __value or self.label == __value

    def __hash__(self) -> int:
        return hash((self.tag, self.label, self.description))

    @classmethod
    def labels(cls) -> List[str]:
        """Get list of labels of all enum members.

        :return: List of all labels
        """
        return [value.label for value in cls.__members__.values()]

    @classmethod
    def tags(cls) -> List[int]:
        """Get list of tags of all enum members.

        :return: List of all tags
        """
        return [value.tag for value in cls.__members__.values()]

    @classmethod
    def contains(cls, obj: Union[int, str]) -> bool:
        """Check if given member with given tag/label exists in enum.

        :param obj: Label or tag of enum
        :return: True if exists False otherwise
        """
        if not isinstance(obj, (int, str)):
            raise SPSDKTypeError("Object must be either string or integer")
        try:
            cls.from_attr(obj)
            return True
        except SPSDKKeyError:
            return False

    @classmethod
    def get_tag(cls, label: str) -> int:
        """Get tag of enum member with given label.

        :param label: Label to be used for searching
        :return: Tag of found enum member
        """
        value = cls.from_label(label)
        return value.tag

    @classmethod
    def get_label(cls, tag: int) -> str:
        """Get label of enum member with given tag.

        :param tag: Tag to be used for searching
        :return: Label of found enum member
        """
        value = cls.from_tag(tag)
        return value.label

    @classmethod
    def get_description(cls, tag: int, default: Optional[str] = None) -> Optional[str]:
        """Get description of enum member with given tag.

        :param tag: Tag to be used for searching
        :param default: Default value if member contains no description
        :return: Description of found enum member
        """
        value = cls.from_tag(tag)
        return value.description or default

    @classmethod
    def from_attr(cls, attribute: Union[int, str]) -> "Self":
        """Get enum member with given tag/label attribute.

        :param attribute: Attribute value of enum member
        :return: Found enum member
        """
        # Let's make MyPy happy, see https://github.com/python/mypy/issues/10740
        if isinstance(attribute, int):
            return cls.from_tag(attribute)
        else:
            return cls.from_label(attribute)

    @classmethod
    def from_tag(cls, tag: int) -> "Self":
        """Get enum member with given tag.

        :param tag: Tag to be used for searching
        :raises SPSDKKeyError: If enum with given label is not found
        :return: Found enum member
        """
        for item in cls.__members__.values():
            if item.tag == tag:
                return item
        raise SPSDKKeyError(
            f"There is no {cls.__name__} item in with tag {tag} defined"
        )

    @classmethod
    def from_label(cls, label: str) -> "Self":
        """Get enum member with given label.

        :param label: Label to be used for searching
        :raises SPSDKKeyError: If enum with given label is not found
        :return: Found enum member
        """
        for item in cls.__members__.values():
            if item.label.upper() == label.upper():
                return item
        raise SPSDKKeyError(
            f"There is no {cls.__name__} item with label {label} defined"
        )


class SpsdkSoftEnum(SpsdkEnum):
    """SPSDK Soft Enum type.

    It has API with default values for labels and
    descriptions with defaults for non existing members.
    """

    @classmethod
    def get_label(cls, tag: int) -> str:
        """Get label of enum member with given tag.

        If member not found and default is specified, the default is returned.

        :param tag: Tag to be used for searching
        :return: Label of found enum member
        """
        try:
            return super().get_label(tag)
        except SPSDKKeyError:
            return f"Unknown ({tag})"

    @classmethod
    def get_description(cls, tag: int, default: Optional[str] = None) -> Optional[str]:
        """Get description of enum member with given tag.

        :param tag: Tag to be used for searching
        :param default: Default value if member contains no description
        :return: Description of found enum member
        """
        try:
            return super().get_description(tag, default)
        except SPSDKKeyError:
            return f"Unknown ({tag})"
